/* (C)opyright MMVI-MMVII Anselm R. Garbe <garbeam at gmail dot com>
 * Copyright (c) 1998 - 2001 Kungliga Tekniska Hgskolan
 * (C)opyright MMVII      Christian Dietrich <stettberger at gmx dot com>
 * See LICENSE file for license details.
 */
#include "dinput.h"
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <paths.h>


void*
emalloc(unsigned int size) {
	void *res = calloc(1, size);

	if(!res)
		eprint("fatal: could not malloc() %u bytes\n", size);
	return res;
}

void
eprint(const char *errstr, ...) {
	va_list ap;

	va_start(ap, errstr);
	vfprintf(stderr, errstr, ap);
	va_end(ap);
	exit(EXIT_FAILURE);
}

char *
estrdup(const char *str) {
	void *res = strdup(str);

	if(!res)
		eprint("fatal: could not malloc() %u bytes\n", strlen(str));
	return res;
}

char *
rline(void)
{
  char *str = emalloc(1024);
  char buf[1024];
  int len = 1024;
  do {
    if (!fgets(buf, sizeof buf, stdin)) {
      if (*str) {
        str[strlen(str)-1] = 0;
        return str;
      }
      else {
        free(str);
        return NULL;
      }
    }
    if (len < strlen(str) + strlen(buf))
      str = erealloc(str, strlen(str) + strlen(buf) + 1);
    strcat(str, buf);
  } while(str[strlen(str)-1] != '\n');
  str[strlen(str)-1] = 0;
  return str;
}

void *
erealloc(void *ptr, size_t size)
{
  void *res = realloc(ptr, size);
  if(!res)
    eprint("fatal: could not realloc %u bytes\n", size);
  return res;
}


#define EX_NOEXEC 126
#define EX_NOTFOUND 127

int
pipe_execv(FILE **stdin_fd, FILE **stdout_fd, FILE **stderr_fd, 
           const char *file, char **argv)
{
  int in_fd[2], out_fd[2], err_fd[2];
  pid_t pid;

  if(stdin_fd != NULL)
    pipe(in_fd);
  if(stdout_fd != NULL)
    pipe(out_fd);
  if(stderr_fd != NULL)
    pipe(err_fd);
  pid = fork();
  switch(pid) {
  case 0:
    if(argv == NULL)
      exit(-1);

    /* close pipes we're not interested in */
    if(stdin_fd != NULL)
      close(in_fd[1]);
    if(stdout_fd != NULL)
      close(out_fd[0]);
    if(stderr_fd != NULL)
      close(err_fd[0]);

    /* pipe everything caller doesn't care about to /dev/null */
    if(stdin_fd == NULL)
      in_fd[0] = open(_PATH_DEVNULL, O_RDONLY);
    if(stdout_fd == NULL)
      out_fd[1] = open(_PATH_DEVNULL, O_WRONLY);
    if(stderr_fd == NULL)
      err_fd[1] = open(_PATH_DEVNULL, O_WRONLY);

    /* move to proper descriptors */
    if(in_fd[0] != STDIN_FILENO) {
      dup2(in_fd[0], STDIN_FILENO);
      close(in_fd[0]);
    }
    if(out_fd[1] != STDOUT_FILENO) {
      dup2(out_fd[1], STDOUT_FILENO);
      close(out_fd[1]);
    }
    if(err_fd[1] != STDERR_FILENO) {
      dup2(err_fd[1], STDERR_FILENO);
      close(err_fd[1]);
    }

    execv(file, argv);
    exit((errno == ENOENT) ? EX_NOTFOUND : EX_NOEXEC);
  case -1:
    if(stdin_fd != NULL) {
      close(in_fd[0]);
      close(in_fd[1]);
    }
    if(stdout_fd != NULL) {
      close(out_fd[0]);
      close(out_fd[1]);
    }
    if(stderr_fd != NULL) {
      close(err_fd[0]);
      close(err_fd[1]);
    }
    return -2;
  default:
    if(stdin_fd != NULL) {
      close(in_fd[0]);
      *stdin_fd = fdopen(in_fd[1], "w");
    }
    if(stdout_fd != NULL) {
      close(out_fd[1]);
      *stdout_fd = fdopen(out_fd[0], "r");
    }
    if(stderr_fd != NULL) {
      close(err_fd[1]);
      *stderr_fd = fdopen(err_fd[0], "r");
    }
  }
  return pid;
}

